@@ -2,7 +2,6 @@
 /* gclue-locator.c
  *
  * Copyright 2013 Red Hat, Inc.
- * Copyright © 2022,2023 Oracle and/or its affiliates.
  *
  * Geoclue is free software; you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free
@@ -19,7 +18,6 @@
  * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  * Authors: Zeeshan Ali (Khattak) <zeeshanak@gnome.org>
- *          Maciej S. Szmigiero <maciej.szmigiero@oracle.com>
  */
 
 #include "config.h"
@@ -28,11 +26,8 @@
 
 #include "gclue-locator.h"
 
-#include "gclue-config.h"
-
-#if GCLUE_USE_WIFI_SOURCE
 #include "gclue-wifi.h"
-#endif
+#include "gclue-config.h"
 
 #if GCLUE_USE_3G_SOURCE
 #include "gclue-3g.h"
@@ -50,37 +45,26 @@
 #include "gclue-nmea-source.h"
 #endif
 
-#if GCLUE_USE_STATIC_SOURCE
-#include "gclue-static-source.h"
-#endif
-
-#if GCLUE_USE_IP_SOURCE
-#include "gclue-ip.h"
-#endif
-
 /* This class is like a master location source that hides all individual
  * location sources from rest of the code
  */
 
-static GClueLocationSourceStartResult
+static gboolean
 gclue_locator_start (GClueLocationSource *source);
-static GClueLocationSourceStopResult
+static gboolean
 gclue_locator_stop (GClueLocationSource *source);
 
+G_DEFINE_TYPE (GClueLocator, gclue_locator, GCLUE_TYPE_LOCATION_SOURCE)
+
 struct _GClueLocatorPrivate
 {
         GList *sources;
         GList *active_sources;
 
         GClueAccuracyLevel accuracy_level;
-        gboolean priority_source_lock;
-        guint64 priority_source_lock_timestamp;
-};
 
-G_DEFINE_TYPE_WITH_CODE (GClueLocator,
-                         gclue_locator,
-                         GCLUE_TYPE_LOCATION_SOURCE,
-                         G_ADD_PRIVATE (GClueLocator))
+        guint time_threshold;
+};
 
 enum
 {
@@ -91,102 +75,40 @@ enum
 
 static GParamSpec *gParamSpecs[LAST_PROP];
 
-#define MAX_SPEED        500       /* Meters per second */
-#define MAX_LOCATION_AGE (30 * 60) /* Seconds. */
-#define MAX_PRIORITY_SOURCE_AGE         30        /* Seconds. */
-#define PRIORITY_ACCURACY_THRESHOLD 20        /* Meters */
-
 static void
 set_location (GClueLocator  *locator,
-              GClueLocationSource *source)
+              GClueLocation *location)
 {
         GClueLocation *cur_location;
-        GClueLocation *location;
-        const char *src_name = NULL;
-        gboolean update_priority_source = FALSE;
-
-        location = gclue_location_source_get_location (source);
-        src_name = G_OBJECT_TYPE_NAME (source);
-
-        if (gclue_location_get_accuracy (location) ==
-            GCLUE_LOCATION_ACCURACY_UNKNOWN) {
-                /* If we do not know the accuracy, discard the update */
-                g_debug ("Discarding %s location with unknown accuracy",
-                         src_name);
-                return;
-        }
+        GeocodeLocation *gloc, *cur_gloc;
 
         cur_location = gclue_location_source_get_location
                         (GCLUE_LOCATION_SOURCE (locator));
+        gloc = GEOCODE_LOCATION (location);
+        cur_gloc = GEOCODE_LOCATION (cur_location);
+
+        g_debug ("New location available");
 
         if (cur_location != NULL) {
-            guint64 cur_timestamp, new_timestamp;
-            double dist, speed;
-
-            cur_timestamp = gclue_location_get_timestamp (cur_location);
-            new_timestamp = gclue_location_get_timestamp (location);
-            if (new_timestamp < cur_timestamp) {
-                    g_debug ("New %s location older than current, ignoring.",
-                             src_name);
+            if (geocode_location_get_timestamp (gloc) <
+                geocode_location_get_timestamp (cur_gloc)) {
+                    g_debug ("New location older than current, ignoring.");
                     return;
             }
 
-            dist = gclue_location_get_distance_from (location, cur_location);
-            if (new_timestamp > cur_timestamp) {
-                guint64 age = new_timestamp - cur_timestamp;
-
-                if (age < MAX_LOCATION_AGE) {
-                    speed = dist / age;
-                } else {
-                    /* The previous location is too old?
-                     * Force the speed to be within the allowed range then.
-                     */
-                    speed = 0;
-                }
-            } else {
-                speed = G_MAXDOUBLE;
-            }
-
-            update_priority_source = gclue_location_source_get_priority_source (source) &&
-                                     (gclue_location_get_accuracy (location) < PRIORITY_ACCURACY_THRESHOLD);
-
-            if (update_priority_source) {
-                     if (!locator->priv->priority_source_lock)
-                              g_debug ("Enabling Priority Source Lock");
-                     locator->priv->priority_source_lock = TRUE;
-                     locator->priv->priority_source_lock_timestamp = new_timestamp;
-            } else if (locator->priv->priority_source_lock &&
-                       (new_timestamp - locator->priv->priority_source_lock_timestamp) >= MAX_PRIORITY_SOURCE_AGE) {
-                     g_debug ("Priority Source Lock no longer active");
-                     locator->priv->priority_source_lock = FALSE;
-            }
-
-            if (locator->priv->priority_source_lock &&
-                !gclue_location_source_get_priority_source (source)) {
-                     g_debug ("Priority Source Lock (age %u s) active, ignoring new %s location",
-                              (guint) (new_timestamp - locator->priv->priority_source_lock_timestamp),
-                              src_name);
-                     return;
-            }
-
-            if (update_priority_source) {
-                     /* A priority source is updating, let it though */
-                     g_debug ("Priority Source Lock Active");
-            } else if ((dist <= gclue_location_get_accuracy (location) ||
-                       speed > MAX_SPEED) &&
-                       gclue_location_get_accuracy (location) >
-                       gclue_location_get_accuracy (cur_location)) {
+            if (geocode_location_get_distance_from (gloc, cur_gloc) * 1000 <
+                geocode_location_get_accuracy (gloc) &&
+                geocode_location_get_accuracy (gloc) >
+                geocode_location_get_accuracy (cur_gloc)) {
                     /* We only take the new location if either the previous one
-                     * lies outside its accuracy circle and was reachable with
-                     * a reasonable speed, OR it is more or as accurate as
-                     * the previous one.
+                     * lies outside its accuracy circle or its more or as
+                     * accurate as previous one.
                      */
-                    g_debug ("Ignoring less accurate new %s location", src_name);
+                    g_debug ("Ignoring less accurate new location");
                     return;
             }
         }
 
-        g_debug ("New location available from %s", src_name);
         gclue_location_source_set_location (GCLUE_LOCATION_SOURCE (locator),
                                             location);
 }
@@ -217,10 +139,8 @@ refresh_available_accuracy_level (GClueLocator *locator)
                         (locator->priv->sources,
                          (GCompareFunc) compare_accuracy_level);
 
-        new = locator->priv->sources ?
-                gclue_location_source_get_available_accuracy_level
-                                (GCLUE_LOCATION_SOURCE (locator->priv->sources->data)) :
-                GCLUE_ACCURACY_LEVEL_NONE;
+        new = gclue_location_source_get_available_accuracy_level
+                        (GCLUE_LOCATION_SOURCE (locator->priv->sources->data));
 
         existing = gclue_location_source_get_available_accuracy_level
                         (GCLUE_LOCATION_SOURCE (locator));
@@ -238,8 +158,10 @@ on_location_changed (GObject    *gobject,
 {
         GClueLocator *locator = GCLUE_LOCATOR (user_data);
         GClueLocationSource *source = GCLUE_LOCATION_SOURCE (gobject);
+        GClueLocation *location;
 
-        set_location (locator, source);
+        location = gclue_location_source_get_location (source);
+        set_location (locator, location);
 }
 
 static gboolean
@@ -262,7 +184,7 @@ start_source (GClueLocator        *locator,
 
         location = gclue_location_source_get_location (src);
         if (gclue_location_source_get_active (src) && location != NULL)
-                set_location (locator, src);
+                set_location (locator, location);
 
         gclue_location_source_start (src);
 }
@@ -378,6 +300,8 @@ gclue_locator_finalize (GObject *gsource)
         GList *node;
         GClueMinUINT *threshold;
 
+        G_OBJECT_CLASS (gclue_locator_parent_class)->finalize (gsource);
+
         threshold = gclue_location_source_get_time_threshold
                         (GCLUE_LOCATION_SOURCE (locator));
         g_signal_handlers_disconnect_by_func
@@ -400,10 +324,7 @@ gclue_locator_finalize (GObject *gsource)
         }
         g_list_free_full (priv->sources, g_object_unref);
         priv->sources = NULL;
-        g_list_free (priv->active_sources);
         priv->active_sources = NULL;
-
-        G_OBJECT_CLASS (gclue_locator_parent_class)->finalize (gsource);
 }
 
 static void
@@ -412,6 +333,7 @@ gclue_locator_constructed (GObject *object)
         GClueLocator *locator = GCLUE_LOCATOR (object);
         GClueLocationSource *submit_source = NULL;
         GClueConfig *gconfig = gclue_config_get_singleton ();
+        GClueWifi *wifi;
         GList *node;
         GClueMinUINT *threshold;
 
@@ -419,7 +341,7 @@ gclue_locator_constructed (GObject *object)
 
 #if GCLUE_USE_3G_SOURCE
         if (gclue_config_get_enable_3g_source (gconfig)) {
-                GClue3G *source = gclue_3g_get_singleton (locator->priv->accuracy_level);
+                GClue3G *source = gclue_3g_get_singleton ();
                 locator->priv->sources = g_list_append (locator->priv->sources,
                                                         source);
         }
@@ -431,21 +353,18 @@ gclue_locator_constructed (GObject *object)
                                                         cdma);
         }
 #endif
-#if GCLUE_USE_WIFI_SOURCE
-        if (gclue_config_get_enable_wifi_source (gconfig)) {
-                GClueWifi *wifi = gclue_wifi_get_singleton (locator->priv->accuracy_level);
-                locator->priv->sources = g_list_append (locator->priv->sources,
-                                                        wifi);
-        }
-#endif
+        if (gclue_config_get_enable_wifi_source (gconfig))
+                wifi = gclue_wifi_get_singleton (locator->priv->accuracy_level);
+        else
+                /* City-level accuracy will give us GeoIP-only source */
+                wifi = gclue_wifi_get_singleton (GCLUE_ACCURACY_LEVEL_CITY);
+        locator->priv->sources = g_list_append (locator->priv->sources, wifi);
 #if GCLUE_USE_MODEM_GPS_SOURCE
         if (gclue_config_get_enable_modem_gps_source (gconfig)) {
                 GClueModemGPS *gps = gclue_modem_gps_get_singleton ();
                 locator->priv->sources = g_list_append (locator->priv->sources,
                                                         gps);
-                if (!submit_source) {
-                        submit_source = GCLUE_LOCATION_SOURCE (gps);
-                }
+                submit_source = GCLUE_LOCATION_SOURCE (gps);
         }
 #endif
 #if GCLUE_USE_NMEA_SOURCE
@@ -453,34 +372,8 @@ gclue_locator_constructed (GObject *object)
                 GClueNMEASource *nmea = gclue_nmea_source_get_singleton ();
                 locator->priv->sources = g_list_append (locator->priv->sources,
                                                         nmea);
-                if (!submit_source) {
-                        submit_source = GCLUE_LOCATION_SOURCE (nmea);
-                }
-
         }
 #endif
-#if GCLUE_USE_STATIC_SOURCE
-        if (gclue_config_get_enable_static_source (gconfig)) {
-                GClueStaticSource *static_source;
-
-                static_source = gclue_static_source_get_singleton
-                        (locator->priv->accuracy_level);
-                locator->priv->sources = g_list_append (locator->priv->sources,
-                                                        static_source);
-        }
-#endif
-#if GCLUE_USE_IP_SOURCE
-        if (gclue_config_get_enable_ip_source (gconfig)) {
-                GClueIp *source = gclue_ip_get_singleton ();
-                locator->priv->sources = g_list_append (locator->priv->sources,
-                                                        source);
-        }
-#endif
-
-        if (locator->priv->sources == NULL) {
-                g_warning ("No sources enabled in configuration, "
-                           "location not available");
-        }
 
         for (node = locator->priv->sources; node != NULL; node = node->next) {
                 g_signal_connect (G_OBJECT (node->data),
@@ -516,6 +409,7 @@ gclue_locator_class_init (GClueLocatorClass *klass)
         object_class->set_property = gclue_locator_set_property;
         object_class->finalize = gclue_locator_finalize;
         object_class->constructed = gclue_locator_constructed;
+        g_type_class_add_private (object_class, sizeof (GClueLocatorPrivate));
 
         gParamSpecs[PROP_ACCURACY_LEVEL] = g_param_spec_enum ("accuracy-level",
                                                               "AccuracyLevel",
@@ -532,27 +426,25 @@ gclue_locator_class_init (GClueLocatorClass *klass)
 static void
 gclue_locator_init (GClueLocator *locator)
 {
-        locator->priv = gclue_locator_get_instance_private (locator);
-        locator->priv->priority_source_lock = FALSE;
-        locator->priv->priority_source_lock_timestamp = 0;
+        locator->priv =
+                G_TYPE_INSTANCE_GET_PRIVATE (locator,
+                                            GCLUE_TYPE_LOCATOR,
+                                            GClueLocatorPrivate);
 }
 
-static GClueLocationSourceStartResult
+static gboolean
 gclue_locator_start (GClueLocationSource *source)
 {
         GClueLocationSourceClass *base_class;
         GClueLocator *locator;
         GList *node;
-        GClueLocationSourceStartResult base_result;
 
-        g_return_val_if_fail (GCLUE_IS_LOCATOR (source),
-                              GCLUE_LOCATION_SOURCE_START_RESULT_FAILED);
+        g_return_val_if_fail (GCLUE_IS_LOCATOR (source), FALSE);
         locator = GCLUE_LOCATOR (source);
 
         base_class = GCLUE_LOCATION_SOURCE_CLASS (gclue_locator_parent_class);
-        base_result = base_class->start (source);
-        if (base_result != GCLUE_LOCATION_SOURCE_START_RESULT_OK)
-                return base_result;
+        if (!base_class->start (source))
+                return FALSE;
 
         for (node = locator->priv->sources; node != NULL; node = node->next) {
                 GClueLocationSource *src = GCLUE_LOCATION_SOURCE (node->data);
@@ -575,26 +467,22 @@ gclue_locator_start (GClueLocationSource *source)
                 start_source (locator, src);
         }
 
-        return base_result;
+        return TRUE;
 }
 
-static GClueLocationSourceStopResult
+static gboolean
 gclue_locator_stop (GClueLocationSource *source)
 {
         GClueLocationSourceClass *base_class;
         GClueLocator *locator;
         GList *node;
-        GClueLocationSourceStopResult base_result;
-
-        g_return_val_if_fail (GCLUE_IS_LOCATOR (source),
-                              GCLUE_LOCATION_SOURCE_STOP_RESULT_FAILED);
 
+        g_return_val_if_fail (GCLUE_IS_LOCATOR (source), FALSE);
         locator = GCLUE_LOCATOR (source);
 
         base_class = GCLUE_LOCATION_SOURCE_CLASS (gclue_locator_parent_class);
-        base_result = base_class->stop (source);
-        if (base_result == GCLUE_LOCATION_SOURCE_STOP_RESULT_STILL_USED)
-                return base_result;
+        if (!base_class->stop (source))
+                return FALSE;
 
         for (node = locator->priv->active_sources; node != NULL; node = node->next) {
                 GClueLocationSource *src = GCLUE_LOCATION_SOURCE (node->data);
@@ -608,7 +496,7 @@ gclue_locator_stop (GClueLocationSource *source)
 
         g_list_free (locator->priv->active_sources);
         locator->priv->active_sources = NULL;
-        return base_result;
+        return TRUE;
 }
 
 GClueLocator *
