@@ -36,12 +36,20 @@ gclue_service_client_client_iface_init (GClueDBusClientIface *iface);
 static void
 gclue_service_client_initable_iface_init (GInitableIface *iface);
 
+G_DEFINE_TYPE_WITH_CODE (GClueServiceClient,
+                         gclue_service_client,
+                         GCLUE_DBUS_TYPE_CLIENT_SKELETON,
+                         G_IMPLEMENT_INTERFACE (GCLUE_DBUS_TYPE_CLIENT,
+                                                gclue_service_client_client_iface_init)
+                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
+                                                gclue_service_client_initable_iface_init));
+
 typedef struct _StartData StartData;
 
 struct _GClueServiceClientPrivate
 {
         GClueClientInfo *client_info;
-        char *path;
+        const char *path;
         GDBusConnection *connection;
         GClueAgent *agent_proxy;
         StartData *pending_auth_start_data;
@@ -49,7 +57,6 @@ struct _GClueServiceClientPrivate
 
         GClueServiceLocation *location;
         GClueServiceLocation *prev_location;
-        GClueLocation *signaled_location;
         guint distance_threshold;
         guint time_threshold;
 
@@ -61,15 +68,6 @@ struct _GClueServiceClientPrivate
         gboolean agent_stopped; /* Agent stopped client, not the app */
 };
 
-G_DEFINE_TYPE_WITH_CODE (GClueServiceClient,
-                         gclue_service_client,
-                         GCLUE_DBUS_TYPE_CLIENT_SKELETON,
-                         G_IMPLEMENT_INTERFACE (GCLUE_DBUS_TYPE_CLIENT,
-                                                gclue_service_client_client_iface_init)
-                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
-                                                gclue_service_client_initable_iface_init)
-                         G_ADD_PRIVATE (GClueServiceClient));
-
 enum
 {
         PROP_0,
@@ -86,13 +84,13 @@ static char *
 next_location_path (GClueServiceClient *client)
 {
         GClueServiceClientPrivate *priv = client->priv;
-        g_autofree char *index_str = NULL;
-        g_autofree char *path = NULL;
+        char *path, *index_str;
 
         index_str = g_strdup_printf ("%u", (priv->locations_updated)++),
         path = g_strjoin ("/", priv->path, "Location", index_str, NULL);
+        g_free (index_str);
 
-        return g_steal_pointer (&path);
+        return path;
 }
 
 /* We don't use the gdbus-codegen provided gclue_client_emit_location_updated()
@@ -125,22 +123,25 @@ distance_below_threshold (GClueServiceClient *client,
                           GClueLocation      *location)
 {
         GClueServiceClientPrivate *priv = client->priv;
+        GClueLocation *cur_location;
         gdouble distance;
-        gdouble threshold;
+        gdouble threshold_km;
 
         if (priv->distance_threshold == 0)
                 return FALSE;
 
-        if (!priv->signaled_location)
-                return FALSE;
+        g_object_get (priv->location,
+                      "location", &cur_location,
+                      NULL);
+        distance = geocode_location_get_distance_from
+                (GEOCODE_LOCATION (cur_location), GEOCODE_LOCATION (location));
+        g_object_unref (cur_location);
 
-        distance = gclue_location_get_distance_from (priv->signaled_location,
-                                                     location);
-        threshold = priv->distance_threshold;
-        if (distance < threshold) {
-                g_debug ("Distance from previous location is %f m and "
-                         "below threshold of %f m.",
-                         distance, threshold);
+        threshold_km = priv->distance_threshold / 1000.0;
+        if (distance < threshold_km) {
+                g_debug ("Distance from previous location is %f km and "
+                         "below threshold of %f km.",
+                         distance, threshold_km); 
                 return TRUE;
         }
 
@@ -152,18 +153,23 @@ time_below_threshold (GClueServiceClient *client,
                       GClueLocation      *location)
 {
         GClueServiceClientPrivate *priv = client->priv;
-        gint64 cur_ts, new_ts;
+        GClueLocation *cur_location;
+        gint64 cur_ts, ts;
         guint64 diff_ts;
 
         if (priv->time_threshold == 0)
                 return FALSE;
 
-        if (!priv->signaled_location)
-                return FALSE;
+        g_object_get (priv->location,
+                      "location", &cur_location,
+                      NULL);
 
-        cur_ts = gclue_location_get_timestamp (priv->signaled_location);
-        new_ts = gclue_location_get_timestamp (location);
-        diff_ts = ABS (new_ts - cur_ts);
+        cur_ts = geocode_location_get_timestamp
+                                (GEOCODE_LOCATION (cur_location));
+        ts = geocode_location_get_timestamp (GEOCODE_LOCATION (location));
+        diff_ts = ABS (ts - cur_ts);
+
+        g_object_unref (cur_location);
 
         if (diff_ts < priv->time_threshold) {
                 g_debug ("Time difference between previous and new location"
@@ -200,19 +206,19 @@ on_locator_location_changed (GObject    *gobject,
         GClueServiceClient *client = GCLUE_SERVICE_CLIENT (user_data);
         GClueServiceClientPrivate *priv = client->priv;
         GClueLocationSource *locator = GCLUE_LOCATION_SOURCE (gobject);
-        GClueLocation *new_location;
-        g_autofree char *path = NULL;
+        GClueLocation *location_info;
+        char *path = NULL;
         const char *prev_path;
-        g_autoptr(GError) error = NULL;
+        GError *error = NULL;
 
-        new_location = gclue_location_source_get_location (locator);
-        if (new_location == NULL)
+        location_info = gclue_location_source_get_location (locator);
+        if (location_info == NULL)
                 return; /* No location found yet */
 
-        if (priv->location != NULL && below_threshold (client, new_location)) {
+        if (priv->location != NULL && below_threshold (client, location_info)) {
                 g_debug ("Updating location, below threshold");
                 g_object_set (priv->location,
-                              "location", new_location,
+                              "location", location_info,
                               NULL);
                 return;
         }
@@ -227,7 +233,7 @@ on_locator_location_changed (GObject    *gobject,
         priv->location = gclue_service_location_new (priv->client_info,
                                                      path,
                                                      priv->connection,
-                                                     new_location,
+                                                     location_info,
                                                      &error);
         if (priv->location == NULL)
                 goto error_out;
@@ -239,16 +245,15 @@ on_locator_location_changed (GObject    *gobject,
 
         gclue_dbus_client_set_location (GCLUE_DBUS_CLIENT (client), path);
 
-        g_clear_object (&priv->signaled_location);
-        priv->signaled_location = g_object_ref (new_location);
-
         if (!emit_location_updated (client, prev_path, path, &error))
                 goto error_out;
-
-        return;
+        goto out;
 
 error_out:
         g_warning ("Failed to update location info: %s", error->message);
+        g_error_free (error);
+out:
+        g_free (path);
 }
 
 static void
@@ -258,11 +263,11 @@ start_client (GClueServiceClient *client, GClueAccuracyLevel accuracy_level)
 
         gclue_dbus_client_set_active (GCLUE_DBUS_CLIENT (client), TRUE);
         priv->locator = gclue_locator_new (accuracy_level);
-        gclue_locator_set_time_threshold (priv->locator, priv->time_threshold);
-        g_signal_connect_object (priv->locator,
-                                 "notify::location",
-                                 G_CALLBACK (on_locator_location_changed),
-                                 client, 0);
+        gclue_locator_set_time_threshold (priv->locator, 0);
+        g_signal_connect (priv->locator,
+                          "notify::location",
+                          G_CALLBACK (on_locator_location_changed),
+                          client);
 
         gclue_location_source_start (GCLUE_LOCATION_SOURCE (priv->locator));
 }
@@ -318,19 +323,21 @@ on_agent_props_changed (GDBusProxy *agent_proxy,
         GVariantIter *iter;
         GVariant *value;
         gchar *key;
-
+        
         if (g_variant_n_children (changed_properties) <= 0)
                 return;
 
         g_variant_get (changed_properties, "a{sv}", &iter);
         while (g_variant_iter_loop (iter, "{&sv}", &key, &value)) {
                 GClueAccuracyLevel max_accuracy;
+                GClueConfig *config;
                 const char *id;
 
                 if (strcmp (key, "MaxAccuracyLevel") != 0)
                         continue;
 
                 gdbus_client = GCLUE_DBUS_CLIENT (client);
+                config = gclue_config_get_singleton ();
                 id = gclue_dbus_client_get_desktop_id (gdbus_client);
                 max_accuracy = g_variant_get_uint32 (value);
                 /* FIXME: We should be handling all values of max accuracy
@@ -347,16 +354,16 @@ on_agent_props_changed (GDBusProxy *agent_proxy,
                         start_client (client, accuracy);
                         g_debug ("Re-started '%s'.", id);
                 } else if (max_accuracy == 0 &&
-                           gclue_dbus_client_get_active (gdbus_client)) {
+                           gclue_dbus_client_get_active (gdbus_client) &&
+                           !gclue_config_is_system_component (config, id)) {
                         stop_client (client);
                         client->priv->agent_stopped = TRUE;
                         g_debug ("Stopped '%s'.", id);
                 }
 
-                g_variant_unref (value);
-                g_variant_iter_free (iter);
                 break;
         }
+        g_variant_iter_free (iter);
 }
 
 struct _StartData
@@ -372,7 +379,7 @@ start_data_free (StartData *data)
 {
         g_object_unref (data->client);
         g_object_unref (data->invocation);
-        g_free (data->desktop_id);
+        g_free(data->desktop_id);
         g_slice_free (StartData, data);
 }
 
@@ -440,7 +447,6 @@ handle_post_agent_check_auth (StartData *data)
         GClueConfig *config;
         GClueAppPerm app_perm;
         guint32 uid;
-        gboolean system_app;
 
         uid = gclue_client_info_get_user_id (priv->client_info);
         max_accuracy = gclue_agent_get_max_accuracy_level (priv->agent_proxy);
@@ -465,11 +471,9 @@ handle_post_agent_check_auth (StartData *data)
         app_perm = gclue_config_get_app_perm (config,
                                               data->desktop_id,
                                               priv->client_info);
-        system_app = (gclue_client_info_get_xdg_id (priv->client_info) == NULL);
 
-        if (app_perm == GCLUE_APP_PERM_ALLOWED || system_app) {
-                /* Since we have no reliable way to identify system apps, no
-                 * need for auth for them. */
+        if (gclue_config_is_system_component (config, data->desktop_id) ||
+            app_perm == GCLUE_APP_PERM_ALLOWED) {
                 complete_start (data);
                 return;
         }
@@ -562,10 +566,9 @@ gclue_service_client_handle_start (GClueDBusClient       *client,
         }
 
         desktop_id = gclue_client_info_get_xdg_id (priv->client_info);
-        if (desktop_id == NULL) {
+        if (desktop_id == NULL)
                 /* Non-xdg app */
                 desktop_id = gclue_dbus_client_get_desktop_id (client);
-        }
 
         if (desktop_id == NULL) {
                 g_dbus_method_invocation_return_error_literal (invocation,
@@ -650,7 +653,6 @@ gclue_service_client_finalize (GObject *object)
         g_clear_object (&priv->locator);
         g_clear_object (&priv->location);
         g_clear_object (&priv->prev_location);
-        g_clear_object (&priv->signaled_location);
         g_clear_object (&priv->client_info);
 
         /* Chain up to the parent class */
@@ -831,9 +833,8 @@ gclue_service_client_handle_set_property (GDBusConnection *connection,
         } else if (ret && strcmp (property_name, "TimeThreshold") == 0) {
                 priv->time_threshold = gclue_dbus_client_get_time_threshold
                         (client);
-                if (GCLUE_IS_LOCATOR (priv->locator))
-                        gclue_locator_set_time_threshold (priv->locator,
-                                                          priv->time_threshold);
+                gclue_locator_set_time_threshold (priv->locator,
+                                                  priv->time_threshold);
                 g_debug ("%s: New time-threshold:  %u",
                          G_OBJECT_TYPE_NAME (client),
                          priv->time_threshold);
@@ -870,6 +871,8 @@ gclue_service_client_class_init (GClueServiceClientClass *klass)
         skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
         skeleton_class->get_vtable = gclue_service_client_get_vtable;
 
+        g_type_class_add_private (object_class, sizeof (GClueServiceClientPrivate));
+
         gParamSpecs[PROP_CLIENT_INFO] = g_param_spec_object ("client-info",
                                                              "ClientInfo",
                                                              "Information on client",
@@ -942,7 +945,9 @@ gclue_service_client_initable_iface_init (GInitableIface *iface)
 static void
 gclue_service_client_init (GClueServiceClient *client)
 {
-        client->priv = gclue_service_client_get_instance_private (client);
+        client->priv = G_TYPE_INSTANCE_GET_PRIVATE (client,
+                                                    GCLUE_TYPE_SERVICE_CLIENT,
+                                                    GClueServiceClientPrivate);
         gclue_dbus_client_set_requested_accuracy_level
                 (GCLUE_DBUS_CLIENT (client), DEFAULT_ACCURACY_LEVEL);
 }
