@@ -31,32 +31,37 @@
 #include "gclue-locator.h"
 #include "gclue-config.h"
 
+/* 20 seconds as milliseconds */
+#define AGENT_WAIT_TIMEOUT 20000
+/*  20 seconds as microseconds */
+#define AGENT_WAIT_TIMEOUT_USEC (20 * G_USEC_PER_SEC)
+
 static void
 gclue_service_manager_manager_iface_init (GClueDBusManagerIface *iface);
 static void
 gclue_service_manager_initable_iface_init (GInitableIface *iface);
 
-struct _GClueServiceManagerPrivate
-{
-        GDBusConnection *connection;
-        GList *clients;
-        GHashTable *agents;
-        GQueue *clients_waiting_agent;
-
-        guint last_client_id;
-        guint num_clients;
-
-        GClueLocator *locator;
-};
-
 G_DEFINE_TYPE_WITH_CODE (GClueServiceManager,
                          gclue_service_manager,
                          GCLUE_DBUS_TYPE_MANAGER_SKELETON,
                          G_IMPLEMENT_INTERFACE (GCLUE_DBUS_TYPE_MANAGER,
                                                 gclue_service_manager_manager_iface_init)
                          G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
-                                                gclue_service_manager_initable_iface_init)
-                         G_ADD_PRIVATE (GClueServiceManager))
+                                                gclue_service_manager_initable_iface_init))
+
+struct _GClueServiceManagerPrivate
+{
+        GDBusConnection *connection;
+        GHashTable *clients;
+        GHashTable *agents;
+
+        guint num_clients;
+        gint64 init_time;
+
+        GClueLocator *locator;
+
+        gboolean active;
+};
 
 enum
 {
@@ -68,29 +73,15 @@ enum
 
 static GParamSpec *gParamSpecs[LAST_PROP];
 
-typedef struct
-{
-        GClueDBusManager *manager;
-        GDBusMethodInvocation *invocation;
-        GClueClientInfo *client_info;
-        gboolean reuse_client;
-
-} OnClientInfoNewReadyData;
-
-static void
-on_client_info_new_ready_data_free (gpointer data)
-{
-        g_slice_free (OnClientInfoNewReadyData, data);
-}
-
 static void
 sync_in_use_property (GClueServiceManager *manager)
 {
-        gboolean in_use = FALSE;
-        GList *l;
+        gboolean in_use = FALSE, active = FALSE;
+        GList *clients, *l;
         GClueDBusManager *gdbus_manager;
 
-        for (l = manager->priv->clients; l != NULL; l = l->next) {
+        clients = g_hash_table_get_values (manager->priv->clients);
+        for (l = clients; l != NULL; l = l->next) {
                 GClueDBusClient *client = GCLUE_DBUS_CLIENT (l->data);
                 GClueConfig *config;
                 const char *id;
@@ -98,6 +89,7 @@ sync_in_use_property (GClueServiceManager *manager)
                 id = gclue_dbus_client_get_desktop_id (client);
                 config = gclue_config_get_singleton ();
 
+                active |= gclue_dbus_client_get_active (client);
                 if (gclue_dbus_client_get_active (client) &&
                     !gclue_config_is_system_component (config, id)) {
                         in_use = TRUE;
@@ -105,74 +97,37 @@ sync_in_use_property (GClueServiceManager *manager)
                         break;
                 }
         }
+        g_list_free (clients);
 
+        if (manager->priv->active != active) {
+                manager->priv->active = active;
+                g_object_notify (G_OBJECT (manager), "active");
+        }
         gdbus_manager = GCLUE_DBUS_MANAGER (manager);
         if (in_use != gclue_dbus_manager_get_in_use (gdbus_manager))
                 gclue_dbus_manager_set_in_use (gdbus_manager, in_use);
 }
 
-static int
-compare_client_bus_name (gconstpointer a,
-                         gconstpointer b)
-{
-        GClueClientInfo *info;
-        const char *bus_name = (const char *) b;
-
-        info = gclue_service_client_get_client_info (GCLUE_SERVICE_CLIENT (a));
-
-        return g_strcmp0 (bus_name, gclue_client_info_get_bus_name (info));
-}
-
-static void
-delete_client (GClueServiceManager *manager,
-                GCompareFunc        compare_func,
-                gpointer            compare_func_data)
-{
-        GClueServiceManagerPrivate *priv = manager->priv;
-        GList *l;
-
-        l = priv->clients;
-        while (l != NULL) {
-                GList *next = l->next;
-
-                if (compare_func (l->data, compare_func_data) == 0) {
-                        g_object_unref (G_OBJECT (l->data));
-                        priv->clients = g_list_delete_link (priv->clients, l);
-                        priv->num_clients--;
-                        if (priv->num_clients == 0) {
-                                g_object_notify (G_OBJECT (manager), "active");
-                        }
-                }
-                l = next;
-        }
-
-        g_debug ("Number of connected clients: %u", priv->num_clients);
-        sync_in_use_property (manager);
-}
-
-static void
-on_client_notify_active (GObject    *gobject,
-                         GParamSpec *pspec,
-                         gpointer    user_data)
-{
-        sync_in_use_property (GCLUE_SERVICE_MANAGER (user_data));
-}
-
 static void
 on_peer_vanished (GClueClientInfo *info,
                   gpointer         user_data)
 {
-        const char *bus_name;
+        GClueServiceManager *manager = GCLUE_SERVICE_MANAGER (user_data);
 
-        bus_name = gclue_client_info_get_bus_name (info);
-        g_debug ("Client `%s` vanished. Dropping associated client objects",
-                 bus_name);
-
-        delete_client (GCLUE_SERVICE_MANAGER (user_data),
-                       compare_client_bus_name,
-                       (char *) bus_name);
+        g_hash_table_remove (manager->priv->clients,
+                             gclue_client_info_get_bus_name (info));
+        g_debug ("Number of connected clients: %u",
+                 g_hash_table_size (manager->priv->clients));
+        sync_in_use_property (manager);
 }
 
+typedef struct
+{
+        GClueDBusManager *manager;
+        GDBusMethodInvocation *invocation;
+        GClueClientInfo *client_info;
+} OnClientInfoNewReadyData;
+
 static gboolean
 complete_get_client (OnClientInfoNewReadyData *data)
 {
@@ -184,34 +139,12 @@ complete_get_client (OnClientInfoNewReadyData *data)
         char *path;
         guint32 user_id;
 
-        /* Disconnect on_peer_vanished_before_completion, if it's there */
-        g_signal_handlers_disconnect_by_data (info, data);
-
         user_id = gclue_client_info_get_user_id (info);
         agent_proxy = g_hash_table_lookup (priv->agents,
                                            GINT_TO_POINTER (user_id));
 
-        if (data->reuse_client) {
-                GList *node;
-                const char *peer;
-
-                peer = g_dbus_method_invocation_get_sender (data->invocation);
-                node = g_list_find_custom (priv->clients,
-                                           peer,
-                                           compare_client_bus_name);
-                if (node != NULL) {
-                        GClueServiceClient *client;
-
-                        client = GCLUE_SERVICE_CLIENT (node->data);
-                        path = g_strdup
-                                (gclue_service_client_get_path (client));
-
-                        goto client_created;
-                }
-        }
-
         path = g_strdup_printf ("/org/freedesktop/GeoClue2/Client/%u",
-                                ++priv->last_client_id);
+                                ++priv->num_clients);
 
         client = gclue_service_client_new (info,
                                            path,
@@ -221,32 +154,24 @@ complete_get_client (OnClientInfoNewReadyData *data)
         if (client == NULL)
                 goto error_out;
 
-        priv->clients = g_list_prepend (priv->clients, client);
-        priv->num_clients++;
-        if (priv->num_clients == 1) {
-                g_object_notify (G_OBJECT (data->manager), "active");
-        }
-        g_debug ("Number of connected clients: %u", priv->num_clients);
-
-        g_signal_connect (client,
-                          "notify::active",
-                          G_CALLBACK (on_client_notify_active),
-                          data->manager);
+        g_hash_table_insert (priv->clients,
+                             g_strdup (gclue_client_info_get_bus_name (info)),
+                             client);
+        g_debug ("Number of connected clients: %u",
+                 g_hash_table_size (priv->clients));
 
         g_signal_connect (info,
                           "peer-vanished",
                           G_CALLBACK (on_peer_vanished),
                           data->manager);
+        g_signal_connect_swapped (client,
+                                  "notify::active",
+                                  G_CALLBACK (sync_in_use_property),
+                                  data->manager);
 
-client_created:
-        if (data->reuse_client)
-                gclue_dbus_manager_complete_get_client (data->manager,
-                                                        data->invocation,
-                                                        path);
-        else
-                gclue_dbus_manager_complete_create_client (data->manager,
-                                                           data->invocation,
-                                                           path);
+        gclue_dbus_manager_complete_get_client (data->manager,
+                                                data->invocation,
+                                                path);
         goto out;
 
 error_out:
@@ -257,36 +182,12 @@ error_out:
 out:
         g_clear_error (&error);
         g_clear_object (&info);
-        on_client_info_new_ready_data_free (data);
+        g_slice_free (OnClientInfoNewReadyData, data);
         g_free (path);
 
         return FALSE;
 }
 
-static void
-on_peer_vanished_before_completion (GClueClientInfo *info,
-                                    gpointer         user_data)
-{
-        OnClientInfoNewReadyData *data = (OnClientInfoNewReadyData *) user_data;
-        GClueServiceManager *self = GCLUE_SERVICE_MANAGER (data->manager);
-        GClueServiceManagerPrivate *priv = self->priv;
-        const char *bus_name;
-
-        bus_name = gclue_client_info_get_bus_name (info);
-        g_debug ("Client `%s` vanished before agent appeared",
-                 bus_name);
-        g_signal_handlers_disconnect_by_data (info,
-                                              user_data);
-        g_dbus_method_invocation_return_error (data->invocation,
-                                               G_DBUS_ERROR,
-                                               G_DBUS_ERROR_DISCONNECTED,
-                                               "%s vanished before completion",
-                                               bus_name);
-        g_queue_remove (priv->clients_waiting_agent, data);
-        on_client_info_new_ready_data_free (data);
-        g_object_unref (info);
-}
-
 static void
 on_client_info_new_ready (GObject      *source_object,
                           GAsyncResult *res,
@@ -298,6 +199,7 @@ on_client_info_new_ready (GObject      *source_object,
         GClueAgent *agent_proxy;
         GError *error = NULL;
         guint32 user_id;
+        gint64 now;
 
         info = gclue_client_info_new_finish (res, &error);
         if (info == NULL) {
@@ -306,7 +208,7 @@ on_client_info_new_ready (GObject      *source_object,
                                                                G_DBUS_ERROR_FAILED,
                                                                error->message);
                 g_error_free (error);
-                on_client_info_new_ready_data_free (data);
+                g_slice_free (OnClientInfoNewReadyData, data);
 
                 return;
         }
@@ -316,109 +218,52 @@ on_client_info_new_ready (GObject      *source_object,
         user_id = gclue_client_info_get_user_id (info);
         agent_proxy = g_hash_table_lookup (priv->agents,
                                            GINT_TO_POINTER (user_id));
-        if (agent_proxy == NULL) {
+        now = g_get_monotonic_time ();
+        if (agent_proxy == NULL &&
+            now < (priv->init_time + AGENT_WAIT_TIMEOUT_USEC)) {
                 /* Its possible that geoclue was just launched on GetClient
                  * call, in which case agents need some time to register
                  * themselves to us.
                  */
-                g_queue_push_tail (priv->clients_waiting_agent, data);
-                g_signal_connect
-                        (info,
-                         "peer-vanished",
-                         G_CALLBACK (on_peer_vanished_before_completion),
-                         user_data);
-                g_debug ("Client `%s` waiting for agent for user ID '%u'",
-                         gclue_client_info_get_bus_name (info), user_id);
-
+                g_timeout_add (AGENT_WAIT_TIMEOUT,
+                               (GSourceFunc) complete_get_client,
+                               user_data);
                 return;
         }
 
         complete_get_client (data);
 }
 
-static void
-handle_get_client (GClueDBusManager      *manager,
-                   GDBusMethodInvocation *invocation,
-                   gboolean               reuse_client)
+static gboolean
+gclue_service_manager_handle_get_client (GClueDBusManager      *manager,
+                                         GDBusMethodInvocation *invocation)
 {
         GClueServiceManager *self = GCLUE_SERVICE_MANAGER (manager);
         GClueServiceManagerPrivate *priv = self->priv;
+        GClueServiceClient *client;
         const char *peer;
         OnClientInfoNewReadyData *data;
 
         peer = g_dbus_method_invocation_get_sender (invocation);
+        client = g_hash_table_lookup (priv->clients, peer);
+        if (client != NULL) {
+                const gchar *existing_path;
+
+                existing_path = gclue_service_client_get_path (client);
+                gclue_dbus_manager_complete_get_client (manager,
+                                                        invocation,
+                                                        existing_path);
+                return TRUE;
+        }
 
         data = g_slice_new (OnClientInfoNewReadyData);
         data->manager = manager;
         data->invocation = invocation;
-        data->reuse_client = reuse_client;
         gclue_client_info_new_async (peer,
                                      priv->connection,
                                      NULL,
                                      on_client_info_new_ready,
                                      data);
-}
-
-static gboolean
-gclue_service_manager_handle_get_client (GClueDBusManager      *manager,
-                                         GDBusMethodInvocation *invocation)
-{
-        handle_get_client (manager, invocation, TRUE);
-
-        return TRUE;
-}
-
-static gboolean
-gclue_service_manager_handle_create_client (GClueDBusManager      *manager,
-                                            GDBusMethodInvocation *invocation)
-{
-        handle_get_client (manager, invocation, FALSE);
-
-        return TRUE;
-}
-
-typedef struct
-{
-        const char *path;
-        const char *bus_name;
-} CompareClientPathNBusNameData;
-
-static int
-compare_client_path_n_bus_name (gconstpointer a,
-                                gconstpointer b)
-{
-        GClueServiceClient *client = GCLUE_SERVICE_CLIENT (a);
-        CompareClientPathNBusNameData *data =
-                (CompareClientPathNBusNameData *) b;
-        GClueClientInfo *info;
-
-        info = gclue_service_client_get_client_info (client);
-
-        if (g_strcmp0 (data->bus_name,
-                       gclue_client_info_get_bus_name (info)) == 0 &&
-            g_strcmp0 (data->path,
-                       gclue_service_client_get_path (client)) == 0)
-                return 0;
-        else
-                return 1;
-}
-
-static gboolean
-gclue_service_manager_handle_delete_client (GClueDBusManager      *manager,
-                                            GDBusMethodInvocation *invocation,
-                                            const char            *path)
-{
-        CompareClientPathNBusNameData data;
-
-        data.path = path;
-        data.bus_name = g_dbus_method_invocation_get_sender (invocation);
-
-        delete_client (GCLUE_SERVICE_MANAGER (manager),
-                       compare_client_path_n_bus_name,
-                       &data);
-
-        gclue_dbus_manager_complete_delete_client (manager, invocation);
-
         return TRUE;
 }
 
@@ -460,7 +305,6 @@ on_agent_proxy_ready (GObject      *source_object,
         guint32 user_id;
         GClueAgent *agent;
         GError *error = NULL;
-        GList *l;
 
         agent = gclue_agent_proxy_new_for_bus_finish (res, &error);
         if (agent == NULL)
@@ -477,19 +321,6 @@ on_agent_proxy_ready (GObject      *source_object,
 
         gclue_dbus_manager_complete_add_agent (data->manager, data->invocation);
 
-        l = priv->clients_waiting_agent->head;
-        while (l != NULL) {
-                GList *next = l->next;
-                OnClientInfoNewReadyData *d =
-                        (OnClientInfoNewReadyData *) l->data;
-
-                if (gclue_client_info_get_user_id(d->client_info) == user_id) {
-                        complete_get_client (d);
-                        g_queue_delete_link (priv->clients_waiting_agent, l);
-                }
-                l = next;
-        }
-
         goto out;
 
 error_out:
@@ -582,16 +413,8 @@ gclue_service_manager_finalize (GObject *object)
 
         g_clear_object (&priv->locator);
         g_clear_object (&priv->connection);
-        if (priv->clients != NULL) {
-                g_list_free_full (priv->clients, g_object_unref);
-                priv->clients = NULL;
-        }
+        g_clear_pointer (&priv->clients, g_hash_table_unref);
         g_clear_pointer (&priv->agents, g_hash_table_unref);
-        if (priv->clients_waiting_agent != NULL) {
-                g_queue_free_full (priv->clients_waiting_agent,
-                                   on_client_info_new_ready_data_free);
-                priv->clients_waiting_agent = NULL;
-        }
 
         /* Chain up to the parent class */
         G_OBJECT_CLASS (gclue_service_manager_parent_class)->finalize (object);
@@ -611,7 +434,7 @@ gclue_service_manager_get_property (GObject    *object,
                 break;
 
         case PROP_ACTIVE:
-                g_value_set_boolean (value, (manager->priv->num_clients != 0));
+                g_value_set_boolean (value, manager->priv->active);
                 break;
 
         default:
@@ -679,6 +502,8 @@ gclue_service_manager_class_init (GClueServiceManagerClass *klass)
         object_class->set_property = gclue_service_manager_set_property;
         object_class->constructed = gclue_service_manager_constructed;
 
+        g_type_class_add_private (object_class, sizeof (GClueServiceManagerPrivate));
+
         gParamSpecs[PROP_CONNECTION] = g_param_spec_object ("connection",
                                                             "Connection",
                                                             "DBus Connection",
@@ -692,8 +517,8 @@ gclue_service_manager_class_init (GClueServiceManagerClass *klass)
         /**
          * GClueServiceManager:active:
          *
-         * Unlike the D-Bus 'InUse' property, this indicates wether or not
-         * currently we have connected clients.
+         * Unlike the D-Bus 'InUse' property, this doesn't differentiate
+         * between system components and apps.
          */
         gParamSpecs[PROP_ACTIVE] = g_param_spec_boolean ("active",
                                                          "Active",
@@ -708,13 +533,19 @@ gclue_service_manager_class_init (GClueServiceManagerClass *klass)
 static void
 gclue_service_manager_init (GClueServiceManager *manager)
 {
-        manager->priv = gclue_service_manager_get_instance_private (manager);
+        manager->priv = G_TYPE_INSTANCE_GET_PRIVATE (manager,
+                                                     GCLUE_TYPE_SERVICE_MANAGER,
+                                                     GClueServiceManagerPrivate);
 
+        manager->priv->clients = g_hash_table_new_full (g_str_hash,
+                                                        g_str_equal,
+                                                        g_free,
+                                                        g_object_unref);
         manager->priv->agents = g_hash_table_new_full (g_direct_hash,
                                                        g_direct_equal,
                                                        NULL,
                                                        g_object_unref);
-        manager->priv->clients_waiting_agent = g_queue_new ();
+        manager->priv->init_time = g_get_monotonic_time ();
 }
 
 static gboolean
@@ -733,10 +564,6 @@ static void
 gclue_service_manager_manager_iface_init (GClueDBusManagerIface *iface)
 {
         iface->handle_get_client = gclue_service_manager_handle_get_client;
-        iface->handle_create_client =
-                gclue_service_manager_handle_create_client;
-        iface->handle_delete_client =
-                gclue_service_manager_handle_delete_client;
         iface->handle_add_agent = gclue_service_manager_handle_add_agent;
 }
 
@@ -760,5 +587,5 @@ gclue_service_manager_new (GDBusConnection *connection,
 gboolean
 gclue_service_manager_get_active (GClueServiceManager *manager)
 {
-        return (manager->priv->num_clients != 0);
+        return manager->priv->active;
 }
